@@ -2,7 +2,7 @@
 
 package HTML::Diff;
 
-$VERSION = '0.54';
+$VERSION = '0.55';
 
 use strict;
 
@@ -23,10 +23,7 @@ use Algorithm::Diff 'sdiff';
 sub member {
     my ($item, @list) = @_;
 
-    foreach my $example (@list) {
-	return 1 if $example eq $item;
-    }
-    return 0;
+    return scalar(grep {$_ eq $item} @list);
 }
 
 sub html_word_diff {
@@ -56,15 +53,18 @@ sub html_word_diff {
 	if ($_ =~ /^<.*>/) {
 	    if ($_ =~ m|^</|) {
 		my ($tag) = m|^</\s*([^ \t\n\r>]*)|;
+		$tag = lc $tag;
+#                print STDERR "Found closer of $tag with " . (scalar @$tagstack) . " stack items\n";
 		# If we found the closer for the tag on top 
 		# of the stack, pop it off.
-		if ($$tagstack[-1] eq $tag) {
+		if ((scalar @$tagstack) > 0 && $$tagstack[-1] eq $tag) {
                     my $stacktag = pop @$tagstack;
                 }
 		return [$_, $tag];
 	    } else {
 		my ($tag) = m|^<\s*([^\s>]*)|;
 		$tag = lc $tag;
+#                print STDERR "Found opener of $tag with " . (scalar @$tagstack) . " stack items\n";
 		if (member($tag, @UNBALANCED_TAGS) || $tag =~ m#/\s*>$#)
 		{	                # (tags without correspond closer tags)
 		    return [$_, $tag];
@@ -86,30 +86,48 @@ sub html_word_diff {
     $tagstack = [];
     @rightchks = map { &$smear_tags } @rightchks;
 
+#    print STDERR Data::Dumper::Dumper(\@leftchks);
+#    print STDERR Data::Dumper::Dumper(\@rightchks);
+	
     # Now do the diff, using the "comparison" half of the pair to
     # compare two chuncks.
     my $chunks = sdiff(\@leftchks, \@rightchks,
 		      sub { $_ = elem_cmprsn(shift); $_ =~ s/\s+$/ /g; $_ });
 
+#    print STDERR Data::Dumper::Dumper($chunks);
+	
     # Finally, process the output of sdiff by concatenating
     # consecutive chunks that were "unchanged."
     my $lastsignal = '';
-    my ($lbuf, $rbuf);
+    my $lbuf = "";
+    my $rbuf = "";
     my @result;
     my $ch;
     foreach $ch (@$chunks) {
 	my ($signal, $left, $right) = @$ch;
-	if ($signal eq 'u' && $lastsignal ne 'u') {
-	    push @result, [$lastsignal, $lbuf, $rbuf];
-	    $lbuf = "";
-	    $rbuf = "";
-	} elsif ($signal ne 'u' && $lastsignal eq 'u') {
-	    push @result, [$lastsignal, $lbuf, $rbuf];
-	    $lbuf = "";
-	    $rbuf = "";
+        if ($signal ne $lastsignal && $lastsignal ne '') {
+	    if ($signal ne 'u' && $lastsignal ne 'u') {
+		$signal = 'c';
+	    } else {
+		push @result, [$lastsignal, $lbuf, $rbuf];
+		$lbuf = "";
+		$rbuf = "";	    
+	    }
 	}
-	$lbuf .= elem_mkp($left) || '';
-	$rbuf .= elem_mkp($right) || '';
+# 	if ($signal eq 'u' && $lastsignal ne 'u') {
+# 	    push @result, [$lastsignal, $lbuf, $rbuf]
+# 		unless $lastsignal eq '';
+# 	    $lbuf = "";
+# 	    $rbuf = "";
+# 	} elsif ($signal ne 'u' && $lastsignal eq 'u') {
+# 	    push @result, [$lastsignal, $lbuf, $rbuf];
+# 	    $lbuf = "";
+# 	    $rbuf = "";
+# 	}
+        my $lelem = elem_mkp($left);
+        my $relem = elem_mkp($right);
+        $lbuf .= (defined $lelem ? $lelem : '');
+        $rbuf .= (defined $relem ? $relem : '');
 	$lastsignal = $signal;
     }
     push @result, [$lastsignal, $lbuf, $rbuf];
@@ -158,38 +176,54 @@ sub word_diff {
 
 =head1 HTML::Diff
 
-This module compares two strings of HTML and returns a list of a chunks
-which indicate the diff between the two input strings, where changes
-in formatting are considered changes.
+This module compares two strings of HTML and returns a list of a
+chunks which indicate the diff between the two input strings, where
+changes in formatting are considered changes.
+
+HTML::Diff does not strictly parse the HTML. Instead, it uses regular
+expressions to make a decent effort at understanding the given HTML.
+As a result, there are many valid HTML documents for which it will not
+produce the correct answer. But there may be some invalid HTML
+documents for which it gives you the answer you're looking for. Your
+mileage may vary; test it on lots of inputs from your domain before
+relying on it.
 
 =head1 SYNOPSIS
 
- $result = html_word_diff($left_text, $right_text);
+    $result = html_word_diff($left_text, $right_text);
 
 =head1 DESCRIPTION
 
-   Returns a reference to a list of triples [<flag>, <left>, <right>].
-   Concatenating all the <left> members from the return value should
-   produce the input $left_text, and likewise for the <right> members.
+Returns a reference to a list of triples [<flag>, <left>, <right>].
+Each triple represents a check of the input texts. The flag tells you
+whether it represents a deletion, insertion, a modification, or an
+unchanged chunk.
 
-   The <flag> is either 'u', '+', '-', or 'c', indicating whether the
-   two chunks are the same, the $right_text contained this chunk and 
-   the left chunk didn't, or vice versa, or the two chunks are simply
-   different. This follows the usage of Algorithm::Diff.
+Every character of each input text is accounted for by some triple in
+the output. Specifically, Concatenating all the <left> members from
+the return value should produce C<$left_text>, and likewise the
+<right> members concatenate together to produce C<$right_text>.
 
-   The difference is computed on a word-by-word basis, "breaking" on
-   visible words in the HTML text. If a tag only is changed, it will
-   not be returned as an independent chunk but will be shown as a
-   change to one of the neighboring words. For balanced tags, such as
-   <b> </b>, it is intended that a change to the tag will be treated
-   as a change to all words in between.
+The <flag> is either C<'u'>, C<'+'>, C<'-'>, or C<'c'>, indicating
+whether the two chunks are the same, the $right_text contained this
+chunk and the left chunk didn't, or vice versa, or the two chunks are
+simply different. This follows the usage of Algorithm::Diff.
+
+The difference is computed on a word-by-word basis, "breaking" on
+visible words in the HTML text. If a tag only is changed, it will not
+be returned as an independent chunk but will be shown as a change to
+one of the neighboring words. For balanced tags, such as <b> </b>, it
+is intended that a change to the tag will be treated as a change to
+all words in between.
 
 =head1 AUTHOR
 
-   Whipped up by Ezra elias kilty Cooper, ezra@ezrakilty.net
+Whipped up by Ezra elias kilty Cooper, <ezra@ezrakilty.net>.
+
+Patch contributed by Adam <asjo@koldfront.dk>.
 
 =head1 SEE ALSO
 
-   Algorithm::Diff
+Algorithm::Diff
 
 =cut
